"""dtool lookup server fixtures"""
import os
import sys

import pytest


# TODO: dtool-lookup-server testing instance provision
@pytest.fixture(scope="session")
def dtool_lookup_server(request):
    pass  # stub
